# -*- coding: utf-8 -*-
from okmodule import Module


class CountLines(Module):
    """给定一个文件计算其行数。
    
    Args:
        infile: <Path>，文件；
        comment_char: <str>，注释字符（以该字符开头的行将被忽略）；
    """
    def __init__(self, infile, comment_char=None):
        self.infile = infile
        self.comment_char = comment_char

    def main(self):
        if self.comment_char is None:
            return sum(1 for _ in open(self.infile))
        return sum(1 for _ in open(self.infile) if not _.startswith(self.comment_char))



# -*- coding: utf-8 -*-
from okmodule import Flag, Option, Argument, Command


class SamtoolsView(Command):
    """使用samtools view处理sam/bam文件。

    需要安装samtools（https://github.com/samtools/samtools）。

    Args:
        bam: Output BAM
        min_mq: Only include in output reads that have mapping quality >= INT
        threads: Number of additional threads to use
        output: Write output to file
        input: Input bam or sam file
    """
    bam = Flag('-b')
    min_mq = Option('-q')
    threads = Option('-@')
    output = Option('-o')
    input = Argument()


class SamtoolsSort(Command):
    """使用samtools sort对bam文件进行排序。

    需要安装samtools（https://github.com/samtools/samtools）。

    Args:
        threads: Number of additional threads to use
        output: Write output to file
        input: Input bam file
    """
    threads = Option('-@')
    output = Option('-o')
    input = Argument()


class SamtoolsIndex(Command):
    """使用samtools index对bam文件生成索引。

    需要安装samtools（https://github.com/samtools/samtools）。

    Args:
        threads: Sets the number of threads
        output: Write index to FILE [alternative to <out.index> in args]
        input: Input bam file
    """
    threads = Option('-@')
    output = Option('-o')
    input = Argument()


class SamtoolsDepth(Command):
    """使用samtools depth计算位点深度。

    需要安装samtools（https://github.com/samtools/samtools）。

    Args:
        all: Output all positions (including zero depth)
        threads: Number of additional threads to use
        output: Write output to FILE
        input: Input bam file
    """
    all_positions = Flag('-a')
    threads = Option('-@')
    output = Option('-o')
    input = Argument()



# -*- coding: utf-8 -*-
from okmodule import Option, Flag, Command


class IQTree(Command):
    """使用iqtree构建进化树

    使用本模块需要安装iqtree3（https://iqtree.github.io）

    Args:
        alignment_file: PHYLIP/FASTA/NEXUS/CLUSTAL/MSF alignment file(s)
        model: Model name
        prefix: Prefix for all output files
        quiet: Quiet mode, suppress printing to screen
        threads: No. cores/threads or AUTO-detect (default: 1)
    """
    alignment_file = Option('-s')
    model = Option('-m')
    prefix = Option('--prefix')
    quiet = Flag('--quiet')
    threads = Option('-T')

    def path(self):
        return ['iqtree']



# -*- coding: utf-8 -*-
from pathlib import Path
from shutil import copyfileobj

from okmodule import Module


class MergeDir(Module):
    """给定一个输入文件夹和一个输出文件，把该文件夹下的所有文件合并到该文件中。

    Args:
        indir: <pathlib.Path>，输入文件夹
        outfile: <pathlib.Path>，输出文件
    """
    def __init__(self, indir, outfile):
        if isinstance(indir, str):
            indir = Path(indir)
        self.indir = indir
        if isinstance(outfile, str):
            outfile = Path(outfile)
        self.outfile = outfile

    def main(self):
        with self.outfile.open('wb') as ofp:
            for infile in self.indir.iterdir():
                if infile.is_dir() or infile.name.startswith('.'):
                    continue
                with open(infile, 'rb') as ifp:
                    copyfileobj(ifp, ofp)



# -*- coding: utf-8 -*-
from okmodule import Option, Command


class Muscle(Command):
    """使用muscle进行多序列比对
    
    使用该模型需要安装muscle（https://github.com/rcedgar/muscle）

    Args:
        align: 需要比对的FASTA文件
        threads: 线程数
        output: 输出文件
    """
    align = Option('-align')
    threads = Option('-threads')
    output = Option('-output')



# -*- coding: utf-8 -*-
from okmodule import Argument, Option, Flag, Command


class Fastqc(Command):
    """使用fastqc进行质量控制。

    使用该模块需要安装fastqc（https://www.bioinformatics.babraham.ac.uk/projects/fastqc/）。

    Args:
        outdir: 输出路径
        threads: 线程数
        extract: 是否解压
        seqfile: 测序文件
    """
    outdir = Option('--outdir')
    threads = Option('--threads')
    extract = Flag('--extract')
    seqfile = Argument()



# -*- coding: utf-8 -*-
from okmodule import Flag, Option, Argument, Command


class Bgzip(Command):
    """bgzip.

    htslib (https://github.com/samtools/htslib) needs to be installed.

    Args:
        force: Overwrite files without asking.
        keep: Don't delete input files during operation.
        threads: Number of compression threads to use.
        file: The file
    """
    force = Flag('-f')
    keep = Flag('-k')
    threads = Option('-@')
    file = Argument()



# -*- coding: utf-8 -*-
import gzip

from okmodule import Module


class Fastq2Fasta(Module):
    """fastq转fasta。
    
    Args:
        fastq: <pathlib.Path>，fastq文件路径
        fasta: <pathlib.Path>，fasta文件路径
    """
    def __init__(self, fastq, fasta):
        self.fastq = fastq
        self.fasta = fasta

    def _open_fastq(self):
        if self.fastq.suffix == '.gz':
            return gzip.open(self.fastq, 'rt')
        return self.fastq.open('r')

    def main(self):
        fastq_fp = self._open_fastq()
        fasta_fp = self.fasta.open('w')
        try:
            while True:
                header = fastq_fp.readline()
                sequence = fastq_fp.readline()
                fastq_fp.readline()
                fastq_fp.readline()
                if not header:
                    break
                if header[0] != '@':
                    continue
                fasta_fp.write(f'>{header[1:]}{sequence}')
        finally:
            fastq_fp.close()
            fasta_fp.close()



# -*- coding: utf-8 -*-
from okmodule import Flag, Option, Argument, Command


class BcftoolsIndex(Command):
    """bcftools index.

    bcftools (https://github.com/samtools/bcftools) need to be installed.

    Args:
        force: Overwrite index if it already exists.
        threads: Use multithreading with INT worker threads.
        vcf: The VCF file.
    """
    force = Flag('-f')
    threads = Option('--threads')
    vcf = Argument()



class BcftoolsConsensus(Command):
    """bcftools consensus.

    bcftools (https://github.com/samtools/bcftools) need to be installed.

    Args:
        fasta_ref: Reference sequence in fasta format.
        mask: Replace regions according to the mask_with option.
        mask_with: Replace with CHAR (skips overlapping variants); change to uppercase (uc) or lowercase (lc).
        output: Write output to a file.
        vcf: The VCF file.
    """
    fasta_ref = Option('-f')
    mask = Option('-m')
    mask_with = Option('--mask-with')
    output = Option('-o')
    vcf = Argument()



# -*- coding: utf-8 -*-
from pathlib import Path

from okmodule import Module

from biomodules.align.muscle import Muscle
from biomodules.phylo.iqtree import IQTree


class MSAPhyloTree(Module):
    """使用多序列比对构建进化树

    多序列比对结果文件和树文件均保存在序列文件同级目录，
    例如，输入文件为/path/to/myseqs.fasta，那么：
    多序列比对文件路径为：/path/to/myseqs.aligned.fasta，
    树文件路径为：/path/to/myseqs.treefile
    
    Args:
        seqs_file: 序列文件
        threads: 线程数，默认为4
    """
    def __init__(self, seqs_file, threads=4):
        if isinstance(seqs_file, str):
            seqs_file = Path(seqs_file)
        self.seqs_file = seqs_file
        self.threads = threads

    def main(self):
        basedir = self.seqs_file.parent
        stem = self.seqs_file.stem
        suffix = self.seqs_file.suffix
        aligned_file = basedir / f'{stem}.aligned{suffix}'
        prefix = str(basedir / stem)
        treefile = basedir / f'{stem}.treefile'
        Muscle(
            align=self.seqs_file,
            threads=self.threads,
            output=aligned_file
        )()
        IQTree(
            alignment_file=aligned_file,
            prefix=prefix,
            quiet=True,
            threads=self.threads
        )()
        return aligned_file, treefile



# -*- coding: utf-8 -*-
from pathlib import Path
from zipfile import ZipFile
from os.path import basename
from shutil import copyfileobj

from okmodule import Module


class MergeZipFile(Module):
    """给定一个zip文件，将该zip文件下的所有文件合并为一个文件

    infile: <pathlib.Path>，zip文件
    outfile: <pathlib.Path>，输出文件
    """
    def __init__(self, infile, outfile):
        if isinstance(infile, str):
            infile = Path(infile)
        self.infile = infile
        if isinstance(outfile, str):
            outfile = Path(outfile)
        self.outfile = outfile

    def main(self):
        with ZipFile(self.infile, 'r') as zip_file, self.outfile.open('wb') as ofp:
            for info in zip_file.infolist():
                if info.is_dir() or basename(info.filename).startswith('.'):
                    continue
                with zip_file.open(info, 'r') as ifp:
                    copyfileobj(ifp, ofp)



# -*- coding: utf-8 -*-
from okmodule import Module


class LowDepth(Module):
    """从samtools depth生成的文件中提取出深度较低的位点。
    
    Args：
        depth_file: samtools depth生成的位点深度文件
        min_depth: 最小深度
        low_depth_file: 低深度位点结果文件
    """
    def __init__(self, depth_file, min_depth, low_depth_file):
        self.depth_file = depth_file
        self.min_depth = min_depth
        self.low_depth_file = low_depth_file

    def main(self):
        with self.depth_file.open() as ifp, self.low_depth_file.open('w') as ofp:
            for line in ifp:
                chrom, pos, depth = line.rstrip().split('\t')
                if int(depth) < self.min_depth:
                    ofp.write(f'{chrom}\t{pos}\n')



# -*- coding: utf-8 -*-
from okmodule import Argument, Option, Command


class Filtlong(Command):
    """使用Filtlong过滤reads。

    使用该模块需要安装Filtlong（https://github.com/rrwick/Filtlong）。
    
    Args:
        min_length: 最小read长度
        min_mean_q: 最小read平均质量值
        input_reads: 输入reads文件路径
    """
    min_length = Option('--min_length')
    min_mean_q = Option('--min_mean_q')
    input_reads = Argument()



# -*- coding: utf-8 -*-
from Bio import SeqIO
from okmodule import Module


class SubSeq(Module):
    """从一个完整的fasta文件中提取部分序列

    Args:
         infile: 完整的fasta文件
            ids: 需要提取的序列编号
        outfile: 输出文件
          extra: 额外内容，像提取的序列文件写入额外信息
    """
    def __init__(self, infile, ids, outfile, extra=None):
        self.infile = infile
        if isinstance(ids, list):
            ids = set(ids)
        self.ids = ids
        self.outfile = outfile
        self.extra = extra

    def main(self):
        with open(self.outfile, 'w') as fp:
            count = 0
            total = len(self.ids)
            for seq in SeqIO.parse(self.infile, 'fasta'):
                if seq.id not in self.ids:
                    continue
                count += 1
                fp.write(seq.format('fasta'))
                if count == total:
                    break
            if self.extra is not None:
                fp.write(self.extra)
